# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from dataclasses import asdict
from typing import Any, Callable, Sequence

from dkg.exceptions import ValidationError
from dkg.manager import DefaultRequestManager
from dkg.method import Method
from dkg.types import TReturn


class Module:
    manager: DefaultRequestManager

    def retrieve_caller_fn(
        self, method: Method[Callable[..., TReturn]]
    ) -> Callable[..., TReturn]:
        def caller(*args: Any, **kwargs: Any) -> TReturn:
            processed_args = method.process_args(*args, **kwargs)
            request_params = asdict(method.action)
            request_params.update(processed_args)

            return self.manager.blocking_request(type(method.action), request_params)

        return caller

    def _attach_modules(self, module_definitions: dict[str, Any]) -> None:
        for module_name, module_info in module_definitions.items():
            module_info_is_list_like = isinstance(module_info, Sequence)

            module = module_info[0] if module_info_is_list_like else module_info

            if hasattr(self, module_name):
                raise AttributeError(
                    f"Cannot set {self} module named '{module_name}'. "
                    " The dkg object already has an attribute with that name"
                )

            setattr(self, module_name, module)

            if module_info_is_list_like:
                if len(module_info) == 2:
                    submodule_definitions = module_info[1]
                    module: "Module" = getattr(self, module_name)
                    module._attach_modules(submodule_definitions)
                elif len(module_info) != 1:
                    raise ValidationError(
                        "Module definitions can only have 1 or 2 elements."
                    )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from dkg.dataclasses import NodeResponseDict
from dkg.manager import DefaultRequestManager
from dkg.method import Method
from dkg.module import Module
from dkg.utils.node_request import NodeRequest
from dkg.utils.blockchain_request import BlockchainRequest
from dkg.types import Address


class Node(Module):
    def __init__(self, manager: DefaultRequestManager):
        self.manager = manager

    _info = Method(NodeRequest.info)

    @property
    def info(self) -> NodeResponseDict:
        return self._info()

    _get_identity_id = Method(BlockchainRequest.get_identity_id)

    def get_identity_id(self, operational: Address) -> int:
        return self._get_identity_id(operational)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from dkg.constants import DefaultParameters
from dkg.dataclasses import BidSuggestionRange
from dkg.manager import DefaultRequestManager
from dkg.method import Method
from dkg.module import Module
from dkg.types import DataHexStr
from dkg.utils.blockchain_request import BlockchainRequest
from dkg.utils.node_request import NodeRequest


class Network(Module):
    def __init__(self, manager: DefaultRequestManager):
        self.manager = manager

    _get_asset_storage_address = Method(BlockchainRequest.get_asset_storage_address)

    _get_bid_suggestion = Method(NodeRequest.bid_suggestion)

    def get_bid_suggestion(
        self,
        public_assertion_id: DataHexStr,
        size_in_bytes: int,
        epochs_number: int,
        range: BidSuggestionRange = BidSuggestionRange.LOW,
    ) -> int:
        content_asset_storage_address = self._get_asset_storage_address(
            "ContentAssetStorage"
        )

        response = self._get_bid_suggestion(
            self.manager.blockchain_provider.blockchain_id,
            epochs_number,
            size_in_bytes,
            content_asset_storage_address,
            public_assertion_id,
            DefaultParameters.HASH_FUNCTION_ID.value,
            range,
        )

        return (
            int(response["bidSuggestion"])
            if range != BidSuggestionRange.ALL
            else response
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from dkg.exceptions import ValidationError
from dkg.types import UAL, Address, ChecksumAddress
from web3 import Web3


def format_ual(
    blockchain: str, contract_address: Address | ChecksumAddress, token_id: int
) -> UAL:
    return f"did:dkg:{blockchain.lower()}/{contract_address.lower()}/{token_id}"


def parse_ual(ual: UAL) -> dict[str, str | Address | int]:
    if not ual.startswith("did:dkg:"):
        raise ValidationError("Invalid UAL!")

    args = ual.replace("did:dkg:", "").split("/")

    if len(args) != 3:
        raise ValidationError("Invalid UAL!")

    blockchain, contract_address, token_id = args

    return {
        "blockchain": blockchain,
        "contract_address": Web3.to_checksum_address(contract_address),
        "token_id": int(token_id),
    }



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


class DKGException(Exception):
    """
    Exception mixin inherited by all exceptions of dkg.py
    This allows::
        try:
            some_call()
        except DKGException:
            # deal with dkg exception
        except:
            # deal with other exceptions
    """


class EnvironmentNotSupported(DKGException):
    """
    Raised when blockchain provider is initialized for unsupported environment.
    """

    pass


class RPCURINotDefined(DKGException):
    """
    Raised when blockchain provider is initialized without RPC URI defined.
    """

    pass


class NetworkNotSupported(DKGException):
    """
    Raised when blockchain provider is initialized for unsupported network.
    """

    pass


class InvalidStateOption(DKGException):
    """
    Raised when invalid state option given to the get operation.
    """

    pass


class MissingKnowledgeAssetState(DKGException):
    """
    Raised when search for the Knowledge Asset state on the network has failed.
    """

    pass


class ValidationError(DKGException):
    """
    Raised when something does not pass a validation check.
    """

    pass


class InvalidRequest(DKGException):
    """
    Raised by a manager to signal that blockchain/node request method
    doesn't exist.
    """

    pass


class HTTPRequestMethodNotSupported(DKGException):
    """
    Raised if used HTTP method isn't supported
    """

    pass


class NodeRequestError(DKGException):
    """
    Raised by Node HTTP Provider if error occurred during request.
    """

    pass


class OperationNotFinished(DKGException):
    """
    Raised when requested operation result isn't ready.
    """

    pass


class OperationFailed(DKGException):
    """
    Raised when requested operation status is failed.
    """

    pass


class AccountMissing(DKGException):
    """
    Raised when trying to perform state-changing blockchain transaction without account
    specified.
    """

    pass


class InvalidDataset(DKGException):
    """
    Raised when dataset URDNA2015 normalization doesn't result in any quads.
    """

    pass


class DatasetInputFormatNotSupported(DKGException):
    """
    Raised when trying to normalize RDF dataset with not supported input format.
    """

    pass


class DatasetOutputFormatNotSupported(DKGException):
    """
    Raised when trying to convert RDF dataset to not supported output format.
    """

    pass


class InvalidKnowledgeAsset(DKGException):
    """
    Raised when root of the Merkle Tree built from N-Quads isn't the same as the
    assertionId.
    """

    pass


class InvalidTokenAmount(DKGException):
    """
    Raised when token amount for operation isn't specified and suggested amount is less
    or equal to what is already present in the contract.
    """

    pass


class LeafNotInTree(DKGException):
    """
    Raised when proof/verification requested for the leaf that is not the part of the
    Merkle Tree.
    """

    pass



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from functools import wraps

from dkg.assertion import Assertion
from dkg.asset import KnowledgeAsset
from dkg.graph import Graph
from dkg.manager import DefaultRequestManager
from dkg.module import Module
from dkg.network import Network
from dkg.node import Node
from dkg.paranet import Paranet
from dkg.providers import BlockchainProvider, NodeHTTPProvider
from dkg.types import UAL, Address, ChecksumAddress
from dkg.utils.ual import format_ual, parse_ual
from dkg.services.input_service import InputService
from dkg.services.node_service import NodeService


class DKG(Module):
    assertion: Assertion
    asset: KnowledgeAsset
    paranet: Paranet
    network: Network
    node: Node
    graph: Graph

    @staticmethod
    @wraps(format_ual)
    def format_ual(
        blockchain: str, contract_address: Address | ChecksumAddress, token_id: int
    ) -> UAL:
        return format_ual(blockchain, contract_address, token_id)

    @staticmethod
    @wraps(parse_ual)
    def parse_ual(ual: UAL) -> dict[str, str | Address | int]:
        return parse_ual(ual)

    def __init__(
        self,
        node_provider: NodeHTTPProvider,
        blockchain_provider: BlockchainProvider,
        config: dict = {},
    ):
        self.manager = DefaultRequestManager(node_provider, blockchain_provider)

        self.initialize_services(self.manager, config)

        modules = {
            "assertion": Assertion(self.manager),
            "asset": KnowledgeAsset(
                self.manager, self.input_service, self.node_service
            ),
            "paranet": Paranet(self.manager),
            "network": Network(self.manager),
            "node": Node(self.manager),
            "graph": Graph(self.manager, self.input_service, self.node_service),
        }
        self._attach_modules(modules)

        # Backwards compatibility
        self.graph.get = self.asset.get.__get__(self.asset)
        self.graph.create = self.asset.create.__get__(self.asset)

    def initialize_services(self, manager, config):
        self.input_service = InputService(manager, config)
        self.node_service = NodeService(manager)

    @property
    def node_provider(self) -> NodeHTTPProvider:
        return self.manager.node_provider

    @node_provider.setter
    def node_provider(self, node_provider: NodeHTTPProvider) -> None:
        self.manager.node_provider = node_provider

    @property
    def blockchain_provider(self) -> BlockchainProvider:
        return self.manager.blockchain_provider

    @blockchain_provider.setter
    def blockchain_provider(self, blockchain_provider: BlockchainProvider) -> None:
        self.manager.blockchain_provider = blockchain_provider



from .general import (
    AutoStrEnum,  # NOQA: F401
    AutoStrEnumCapitalize,  # NOQA: F401
    AutoStrEnumUpperCase,  # NOQA: F401
)
from .blockchain import (
    ABI,  # NOQA: F401
    ABIElement,  # NOQA: F401
    ABIError,  # NOQA: F401
    ABIEvent,  # NOQA: F401
    ABIFunction,  # NOQA: F401
    ABIParameter,  # NOQA: F401
    AgreementData,  # NOQA: F401
    Environment,  # NOQA: F401
)
from .dkg_node import UAL  # NOQA: F401
from .encoding import BytesLike, DataHexStr, HexStr  # NOQA: F401
from .evm import Address, ChecksumAddress, Wei  # NOQA: F401
from .generics import TFunc, TReturn  # NOQA: F401
from .network import URI  # NOQA: F401
from .rdf import JSONLD, NQuads  # NOQA: F401
